import numpy as np
import pandas as pd
import seaborn as sns
import warnings
import matplotlib.pyplot as plt

def birds_eye_view(df, n=20, var_list=None):
    """Takes in a pandas.DataFrame object, an optional integer for the histogram bin size, an optional custom variable list, and displays 3 different visualization sets.

    1. Histograms for each numeric variable
    2. A bar chart for each categorical variable
    3. A correlation heatmap of the numeric variables.

    Parameters
    ----------
    df : pandas.DataFrame
        dataframe to create the visualizations
    n : int
        bin size for histograms
    var_list : list
        a specific list of variables to examine, defaults to None

    Returns
    -------
    charts: dict
        A dictionary containing the plot objects created by this function

    Examples
    --------
    >>> birds_eye_view(df, n=30)

    """

    if not (type(df) == pd.DataFrame):
        raise TypeError("df must be input as a DataFrame.")

    if type(var_list) != list and var_list is not None:
        raise TypeError("var_list must be a list.")

    if type(n) != int:
        raise TypeError("n must be an integer.")

    # Generate the visualizations

    viz = {}
    heatmap_list = []
    heatmap = []
    histograms = []
    bar_charts = []

    # Defining the numeric and categorical variables
    numeric = df.select_dtypes(include=np.number).columns.tolist()
    categorical = df.select_dtypes(include=("object" or "string")).columns.to_list()

    # Plot all the variables
    if var_list is None:

        # Histograms
        for num_col in numeric:
            chart = sns.histplot(df, x=(num_col), bins=n, kde=True)
            plt.title("Histogram for " + num_col)
            plt.figure()
            histograms.append(chart)

        # Bar Charts
        for cat_col in categorical:
            if len(pd.unique(df[cat_col])) > 11:
                print(cat_col, " has too many unique values")
            else:
                chart = sns.countplot(data=df, x=(cat_col))
                plt.title("Bar Chart for " + cat_col)
                plt.figure()
                bar_charts.append(chart)

        # Heatmap
        corr_matrix = df[numeric].corr()
        mask = np.triu(np.ones_like(corr_matrix, dtype=np.bool_))
        chart = sns.heatmap(data=corr_matrix,
                            vmin=-1,
                            vmax=1,
                            annot=True,
                            cmap="BrBG",
                            mask=mask
        )
        plt.title("Heatmap of correlation between numeric features")
        plt.figure(figsize=(12, 6))
        print(chart)
        viz["heatmap"] = chart

    # Plot just the custom variables from var_list (if applicable)
    else:
        
        for custom_col in var_list:
            all_cols = df.columns.to_list()
            if custom_col not in all_cols:
                raise TypeError("Variable name " +
                                custom_col +
                                " not found in data frame, please check inputs in var_list.")

            # Histograms
            if custom_col in numeric:
                heatmap_list.append(custom_col)
                chart = sns.histplot(df, x=(custom_col), bins=n, kde=True)
                plt.title("Histogram for " + custom_col)
                plt.figure()
                histograms.append(chart)

            # Bar Charts
            elif custom_col in categorical:
                if len(pd.unique(df[custom_col])) > 11:
                    print(custom_col, " has too many unique values")
                else:
                    chart = sns.countplot(data=df, x=(custom_col))
                    plt.title("Bar Chart for " + custom_col)
                    plt.figure()
                    bar_charts.append(chart)

        # Heatmap
        corr_matrix = df[heatmap_list].corr()
        mask = np.triu(np.ones_like(corr_matrix, dtype=np.bool_))
        chart = sns.heatmap(data=corr_matrix,
                            vmin=-1,
                            vmax=1,
                            annot=True,
                            cmap="BrBG",
                            mask=mask
        )
        plt.title("Heatmap of correlation between numeric features")
        plt.figure(figsize=(12, 6))
        viz["heatmap"] = chart

    viz["histograms"] = histograms
    viz["bar_charts"] = bar_charts
    
    return viz


import pandas as pd
import numpy as np
import altair as alt


def close_up(df, n=1):
    """Accepts a dataframe and the number of pairs of variables with strongest correlations, and
    returns vertically combined scatterplots with a correlation trend for each pair. 

        Parameters
        ----------
        df : pd.core.frame.DataFrame
             dataframe to create the visualization(s)
        n : int
            number of pairs of variables with strongest correlations to be displayed,
            defaults to 1
            
        Returns
        -------
        chart: altair.vegalite.v4.api.Chart (if n = 1)
               altair.vegalite.v4.api.VConcatChart (if n > 1)
               Vertically combined scatterplots with a correlation trend for each pair

        Examples
        --------
        >>> close_up(df, n = 4)
    """

    # check if input is a DataFrame
    if not isinstance(df, pd.core.frame.DataFrame):
        raise TypeError("df should be of type 'pandas.core.frame.DataFrame'")

    if not isinstance(n, int):
        raise TypeError("n should be of type 'int'.")

    # calculate max allowable integer
    numeric = df.select_dtypes(include=np.number).columns.tolist()
    corr_matrix = df[numeric].corr()
    N_max = len(corr_matrix) * (len(corr_matrix) - 1) / 2

    # check if input exceeds max allowable integer
    if n > N_max:
        raise ValueError("n exceeds total number of coefficients.")

    # initialization
    corr = corr_matrix.to_numpy()
    corr = abs(corr)
    corr = np.triu(corr)  # take only upper triangle
    np.fill_diagonal(corr, 0)  # zero diagonal
    max_row = np.argmax(np.max(corr, axis=1))
    max_col = np.argmax(np.max(corr, axis=0))
    viz = {}  # viz dict to be returned

    # plot
    for i in range(n):
        max_row = np.argmax(np.max(corr, axis=1))
        max_col = np.argmax(np.max(corr, axis=0))
        coef = corr_matrix.iloc[max_row, max_col]
        points = (
            alt.Chart(df, title=f'coeff: {coef:.3f}')
            .mark_point(opacity=0.3)
            .encode(
                alt.X(corr_matrix.columns[max_row]), alt.Y(
                    corr_matrix.columns[max_col])
            )
        )
        viz[i+1] = points + points.transform_regression(
            corr_matrix.columns[max_row], corr_matrix.columns[max_col]).mark_line(size=3)
        corr[max_row, max_col] = 0

    # generate one big chart
    chart = viz[1]
    for i in range(n-1):
        chart &= viz[i+2]

    return chart



import numpy as np
import pandas as pd
from scipy import stats

def clean_up(df):
    """Takes a dataframe object and returns a cleaned version 
     with rows containing any NaN values dropped. 
     Inspects the clean dataframe and prints a list of potential outliers for each explanatory variable, 
     based on the threshold distance of 3 standard deviations.
        Parameters
        ----------
        df : dataframe
            dataframe to be cleaned
    
        Returns
        -------
        df_clean
            same dataframe with all the NaN's removed
        Examples
        --------
        >>> df_clean = clean_up(df)
                
        '**The following potenital outliers were detected:**
        Variable X: 
        [ 300, 301, 500, 1000 ]
        Variable Y: 
        [ 6.42, 6.44, 58.52, 60.22 ]'
    
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError("the input df must be pd.DataFrame type")
    
    # Drop any row that contains missing value and reset the index
    df_clean = df.dropna(axis=0, how='any').reset_index(drop=True)

    # Keep only numerical variables relevant for outlier detection
    num_df = df_clean.select_dtypes(['number'])
    
    # Create a dataframe that contains only the outliers
    outlier_df = num_df[(np.abs(stats.zscore(num_df)) > 3)] 

    # Prints out unique outlier values for each variable
    print("**The following potenital outliers were detected:**")
    for col in outlier_df:
        outliers = outlier_df[col].dropna()
        if len(outliers) != 0:
            print(f"Variable {col}: ")
            print(np.unique(outliers.values))

    # returns the clean dataframe with NaN values dropped
    return df_clean


import pandas as pd
import numpy as np

def summary_suggestions(df, threshold = 0.8):
    """Takes in a pandas dataframe and returns a list object comprising
    of 3 dataframes and a list. The dataframes correspond to the
    summary statistics of numeric and categorical variables each and
    the proportion of unique values for categorical variables. The 
    nested list is of the categorical variables that exceed the threshold
    for considering dropping variables with high unique values.

    Parameters
    ----------
    df : pandas dataframe
        Dataframe to be examined

    threshold : float
        threshold for considering dropping variables with high unique values

    Returns
    -------
    results : list
        List of summary dataframes

    Examples
    --------
    >>> summary_suggestions(df)

    [
    (summary statistics for numeric variables),
    (summary statistics for categorical variables),
    (percentage of unique values for categorical variables),
    [list of variables with percentage of unique values higher than the threshold]
    ]
    
    """

    # check if input is a DataFrame
    if not isinstance(df, pd.DataFrame):
        raise TypeError("Input df must be a pandas dataframe object")

    if not ((type(threshold) == float) | (type(threshold) == int)):
        raise TypeError("Input threshold must be a float value between 0 and 1")

    if not (0 < threshold < 1):
        raise TypeError("Input threshold must be a float value between 0 and 1")

    numeric_summary_df = df.select_dtypes(include=np.number).describe()
    categorical_summary_df = df.select_dtypes(include=np.object_).describe()

    results = []
    results.extend([numeric_summary_df, categorical_summary_df])

    unique_val_df = categorical_summary_df[categorical_summary_df.index == 'unique']/len(df)
    filtered_unique_val_df = unique_val_df.loc['unique'] > threshold
    unique_val_vars = [*filter(filtered_unique_val_df.get, filtered_unique_val_df.index)]

    results.extend([unique_val_df, unique_val_vars])
    return results
    


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyeasyeda']

package_data = \
{'': ['*']}

install_requires = \
['altair>=4.2.0,<5.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'seaborn>=0.11.2,<0.12.0']

setup_kwargs = {
    'name': 'pyeasyeda',
    'version': '0.2.5',
    'description': 'Make exploratory data analysis easier! ',
    'long_description': "# pyeasyeda\n\n[![ci-cd](https://github.com/UBC-MDS/pyeasyeda/actions/workflows/ci-cd.yml/badge.svg)](https://github.com/UBC-MDS/pyeasyeda/actions/workflows/ci-cd.yml) [![codecov](https://codecov.io/gh/UBC-MDS/pyeasyeda/branch/main/graph/badge.svg?token=vaOyqFqkor)](https://codecov.io/gh/UBC-MDS/pyeasyeda)\n\nSince exploratory data analysis is an imperative part of every analysis, this package aims at providing efficient data scrubbing and visualization tools to perform preliminary EDA on raw data. The package can be leveraged to clean the dataset and visualize relationships between features to generate insightful trends.\n\n## Functions\n\n-   `clean_up` - This function takes in a pandas dataframe object and performs initial steps of EDA on unstructured data. It returns a clean dataset by removing null values and identifying potential outliers in numeric variables based on a defined threshold.\n\n-   `birds_eye_view` - This function takes in a pandas dataframe object and visualizes the distributions of variables in the form of histograms and density plots. It also generates a correlation heatmap for numeric variables to study their relationships.\n\n-   `close_up` - This function accepts a pandas dataframe object creates a scatterplot of the variable(s) most strongly correlated with the dependent variable. The plot also produces a trend line to model the correlation between the variables.\n\n-   `summary_suggestions` - This function takes in a pandas dataframe object and outputs a table of summary statistics for numeric and categorical variables and a table for percentage of unique values in the categorical variables.\n\nOther packages that offer similar functionality are:\n- [datascience_eda](https://github.com/UBC-MDS/datascience_eda)\n- [QuickDA](https://github.com/sid-the-coder/QuickDA)\n- [easy-data-analysis](https://github.com/jschnab/easy-data-analysis)\n\n## Installation\n\n```bash\n$ pip install pyeasyeda\n```\n\n## Usage\nPlease refer to the documentation link provided below, under 'Example usage' section, for the detailed demonstration of how to use the package.\n\n## Documentation\nThe official documentation is hosted on Read the Docs: https://pyeasyeda.readthedocs.io/en/latest/\n\n## Contributors\nThis python package was developed by James Kim, Kristin Bunyan, Luming Yang and Sukhleen Kaur. The team is from the Master of Data Science program at the University of the British Columbia.\n\n## Contributing\n\nInterested in contributing? Check out the contributing guidelines. Please note that this project is released with a Code of Conduct. By contributing to this project, you agree to abide by its terms.\n\n## License\n\n`pyeasyeda` was created by James Kim, Kristin Banyan, Luming Yang and Sukhleen Kaur. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`pyeasyeda` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n",
    'author': 'Luming Yang',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
